"""
By default, Python's logging library is not safe to use with multiprocessing. Simply logging to a
file or the console can result in messages becoming interleaved or other unexpected behavior. It may
be desirable for plugins to log to the same handler as the host process instead of having their own
log files or output streams. One way of accomplishing this is for plugins to push their log messages
to a queue which is serviced by a single process.
"""

import logging
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import Collection


def configure_child_process_logger(ipc_logger_queue: Queue, level: int = logging.NOTSET):
    """
    Configures a child process to send all log messages to a queue

    This function can be run on a child process to configure it to push all its log messages to a
    queue. Note that formatting of the log messages will be left to the process servicing the queue.
    Additionally, the only filtering performed is based on log level. By default, all log messages
    are pushed to the queue.

    :param ipc_logger_queue: A Queue object shared between the plugin and the queue servicing
                             process
    :param level: The minimum log level of statements to push to the queue; messages below this log
                  level will be dropped, defaults to logging.NOTSET
    """
    root = logging.getLogger()
    root.addHandler(QueueHandler(ipc_logger_queue))
    root.setLevel(level)


def configure_host_process_logger(
    ipc_logger_queue: Queue,
    handlers: Collection[logging.Handler] = [],
) -> QueueListener:
    """
    Configures the root logger to use a QueueListener

    A QueueListener can be used to process the log messages from a child process. This function
    configures a QueueListener to use the provided `ipc_logger_queue` and handlers. It configures
    the root logger to push log messages from the host process into the `ipc_logger_queue`. Finally,
    it returns the QueueListener.

    Note that you will need to call `QueueListener.start()`, otherwise the log messages will not be
    processed. See https://docs.python.org/3/library/logging.handlers.html#queuelistener for more
    information about QueueListener

    :param ipc_logger_queue: A Queue shared by the host and child process that stores log messages
    :param handlers: A Collection of LogHandler objects that the QueueListener will use to handle
                     log messages it pulls off of the ipc_logger_queue

    :return: An unstarted QueueListener object
    """
    root = logging.getLogger()
    root.addHandler(logging.handlers.QueueHandler(ipc_logger_queue))

    return QueueListener(ipc_logger_queue, *handlers, respect_handler_level=True)



from pathlib import Path
from typing import Optional, Union

from . import MultiprocessingPlugin, MultiUsePlugin, PluginThreadName
from .nop import NOP
from .plugin_wrapper import PluginWrapper
from .types import ConfigureLoggerCallback as ConfigureLoggerCallback


class PluginLoader:
    """
    Loads plugins from the provided plugin directory
    """

    def __init__(
        self, plugin_directory: Path, configure_child_process_logger: ConfigureLoggerCallback = NOP
    ):
        """
        :param plugin_directory: The directory where plugins are stored
        :param configure_child_process_logger: A callback to configure logging that will be run
                                               by any MultiprocessingPlugin that this object loads.
                                               This can be overridden on each call to
                                               load_multiprocessing_plugin(). Defaults to a NOP.
        """
        self._plugin_directory = plugin_directory
        self._configure_child_process_logger = configure_child_process_logger

    def load(
        self, *, plugin_name: str, reset_modules_cache: bool = True, **kwargs
    ) -> MultiUsePlugin:
        """
        Load a plugin by name

        Reseting `sys.modules` does not work properly if extension modules have been loaded. The
        `reset_modules_cache` option gives the caller the ability to disable this behavior.

        :param plugin_name: The name of the plugin (corresponds to the name of the directory where
                            the plugin is stored)
        :param reset_modules_cache: Whether or not to reset the `sys.modules` cache to system
                                    defaults before executing the plugin. Setting this to `False`
                                    will break plugin isolation for these plugins. Defaults to
                                    `True`.
        :param kwargs: Keyword arguments to be passed to the plugin's constructor

        :return: A MultiUsePlugin
        """
        return PluginWrapper(
            plugin_name=plugin_name,
            plugin_directory=self._plugin_directory / plugin_name,
            reset_modules_cache=reset_modules_cache,
            **kwargs,
        )

    def load_multiprocessing_plugin(
        self,
        *,
        plugin_name: str,
        main_thread_name: Union[PluginThreadName, str] = PluginThreadName.DEFAULT,
        configure_child_process_logger: Optional[ConfigureLoggerCallback] = None,
        reset_modules_cache=True,
        **kwargs,
    ) -> MultiprocessingPlugin:
        """
        Load a MultiprocessingPlugin by name

        Reseting `sys.modules` does not work properly if extension modules have been loaded. The
        `reset_modules_cache` option gives the caller the ability to disable this behavior.

        :param plugin_name: The name of the plugin (corresponds to the name of the directory where
                            the plugin is stored)
        :param main_thread_name: The name of the child process's main thread. This can either be a
                                 `PluginThreadName` or a string. If it is
                                 `PluginThreadName.DEFAULT`, then child process's main thread will
                                 be whatever the interpreter's default main thread name is. If it is
                                 `PluginThreadName.CALLING_THREAD`, then the child process's main
                                 thread name will match the name of the thread that calls `run()` on
                                 this plugin. If it is a string, the child process's main thread
                                 name will be set to the string value.

                                 Setting this is useful when analyzing logs for applications that
                                 are both multi-threaded and use `MultiprocessingPlugins`, defaults
                                 to `PluginThreadName.DEFAULT`.
        :param configure_child_process_logger: A callback to configure logging on the child process.
                                               This overrides the callback provided to the
                                               constructor. Defaults to `None`
        :param reset_modules_cache: Whether or not to reset the `sys.modules` cache to system
                                    defaults before executing the plugin. Setting this to `False`
                                    will have little to no effect in most cases since
                                    `MultiprocessingPlugins` use the "spawn" method. Defaults to
                                    `True`.
        :param kwargs: Keyword arguments to be passed to the plugin's constructor

        :return: A MultiprocessingPlugin
        """
        plugin = PluginWrapper(
            plugin_name=plugin_name,
            plugin_directory=self._plugin_directory / plugin_name,
            reset_modules_cache=reset_modules_cache,
            **kwargs,
        )

        if configure_child_process_logger is None:
            configure_logger_fn = self._configure_child_process_logger
        else:
            configure_logger_fn = configure_child_process_logger

        return MultiprocessingPlugin(
            plugin=plugin,
            main_thread_name=main_thread_name,
            configure_child_process_logger=configure_logger_fn,
            **kwargs,
        )



import contextlib
import importlib
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Optional

from . import CLEAN_SYS_MODULES, MultiUsePlugin, NamedPluginMixin
from .constants import VENDOR_DIRECTORY_NAME


class PluginWrapper(NamedPluginMixin, MultiUsePlugin):
    """
    Wraps a Plugin to make loading and import system manipulation transparent

    Plugins are isolated by manipulating the import system. This needs to be performed just before
    the plugin is run and the import system must be restored to how it was before it was manipulated
    by this component and the plugin.
    """

    def __init__(
        self,
        *,
        plugin_name: str,
        plugin_directory: Path,
        reset_modules_cache: bool = True,
        **kwargs,
    ):
        super().__init__(plugin_name=plugin_name)

        self._plugin_directory = plugin_directory
        self._vendor_directory = self._plugin_directory / VENDOR_DIRECTORY_NAME
        self.plugin: Optional[MultiUsePlugin] = None
        self._reset_modules_cache = reset_modules_cache

        self._constructor_kwargs = kwargs

    def run(self, **kwargs) -> Any:
        # After a plugin has been imported and run, it contains references to its own
        # modules/dependencies. Therefore, we don't need to use `self._plugin_import_context()` in
        # order to isolate after the initial run.
        if self.plugin is not None:
            return self.plugin.run(**kwargs)

        return self._run_in_isolated_context(**kwargs)

    def _run_in_isolated_context(self, **kwargs) -> Any:
        result = None
        exception = None

        with self._plugin_import_context():
            try:
                self.plugin = self._load_plugin()
                result = self.plugin.run(**kwargs)
            except Exception as ex:
                exception = ex

        if exception is not None:
            raise exception

        return result

    @contextlib.contextmanager
    def _plugin_import_context(self):
        """
        This context manager performs the following:

        1. Save the state of sys.modules
        2. Reset sys.modules to the interpreter's defaults
        3. Configure the import system to import the plugin and its dependencies
        4. yield
        5. Restore the state of the import system.
        """

        if self._reset_modules_cache:
            with self._clean_system_modules():
                with self._plugin_import_path():
                    yield
        else:
            with self._plugin_import_path():
                yield

    @contextlib.contextmanager
    def _clean_system_modules(self):
        host_process_sys_modules = sys.modules.copy()
        PluginWrapper._set_sys_modules(CLEAN_SYS_MODULES)

        yield

        PluginWrapper._set_sys_modules(host_process_sys_modules)

    @contextlib.contextmanager
    def _plugin_import_path(self):
        sys.path = [str(self._plugin_directory.parent), str(self._vendor_directory), *sys.path]

        yield

        sys.path = sys.path[2:]

    @staticmethod
    def _set_sys_modules(modules: Dict[str, ModuleType]):
        # WARNING: Attempting to set sys.modules like `sys.modules = modules` may cause Python to
        #          fail. See https://docs.python.org/3/library/sys.html#sys.modules
        sys.modules.clear()
        sys.modules.update(modules)

    def _load_plugin(self) -> MultiUsePlugin:
        plugin_class = importlib.import_module(f"{self.name}.plugin").Plugin
        return plugin_class(plugin_name=self.name, **self._constructor_kwargs)



from typing import Any

from typing_extensions import Protocol


class SingleUsePlugin(Protocol):
    """A protocol for a plugin that can be run only once"""

    def run(self, **kwargs) -> Any:
        """
        Run a plugin with the provided keyword arguments and returns the result

        A SingleUsePlugin's run() method can be called once and only once. After the plugin's run()
        method returns, the plugin object should be discarded.
        """
        pass

    @property
    def name(self) -> str:
        """A name that identifies this plugin"""


class MultiUsePlugin(SingleUsePlugin, Protocol):
    """A protocol for plugins"""

    def run(self, **kwargs) -> Any:
        """
        Run a plugin with the provided keyword arguments and returns the result

        Unlike a SingleUsePlugin, a MultiUsePlugin's run() method can be called repeatedly with no
        ill effects.
        """
        pass

    @property
    def name(self) -> str:
        """A name that identifies this plugin"""



from types import TracebackType
from typing import Optional, Type

from typing_extensions import Protocol


class Lock(Protocol):
    def __enter__(self) -> bool:
        ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        ...

    def acquire(self, blocking: bool = ..., timeout: float = ...) -> bool:
        ...

    def release(self) -> None:
        ...

    def locked(self) -> bool:
        ...


class Event(Protocol):
    def is_set(self) -> bool:
        ...

    def set(self) -> None:
        ...

    def clear(self) -> None:
        ...

    def wait(self, timeout: Optional[float] = ...) -> bool:
        ...



# This needs to be the first thing imported in the host program so that we can save the state of the
# import system before it gets modified.
import sys

CLEAN_SYS_MODULES = sys.modules.copy()

from . import concurrency
from . import types
from . import logging
from .plugin_thread_name import PluginThreadName
from .plugin import SingleUsePlugin, MultiUsePlugin
from .named_plugin_mixin import NamedPluginMixin
from .multiprocessing_plugin import MultiprocessingPlugin
from .plugin_loader import PluginLoader



import logging
import multiprocessing
from threading import current_thread
from typing import Any, Optional, Union

from . import NamedPluginMixin, PluginThreadName, SingleUsePlugin
from .constants import SERPENTARIUM
from .nop import NOP
from .types import ConfigureLoggerCallback as ConfigureLoggerCallback

logger = logging.getLogger(SERPENTARIUM)


class MultiprocessingPlugin(NamedPluginMixin, SingleUsePlugin):
    """
    A plugin that runs concurrently in a separate process
    """

    def __init__(
        self,
        *,
        plugin: SingleUsePlugin,
        main_thread_name: Union[PluginThreadName, str] = PluginThreadName.DEFAULT,
        daemon: bool = False,
        configure_child_process_logger: ConfigureLoggerCallback = NOP,
        **kwargs,
    ):
        """
        :param plugin: A Plugin to run in a separate process
        :param main_thread_name: The name of the child process's main thread. This can either be a
                                 `PluginThreadName` or a string. If it is
                                 `PluginThreadName.DEFAULT`, then child process's main thread will
                                 be whatever the interpreter's default main thread name is. If it is
                                 `PluginThreadName.CALLING_THREAD`, then the child process's main
                                 thread name will match the name of the thread that calls `run()` on
                                 this plugin. If it is a string, the child process's main thread
                                 name will be set to the string value.

                                 Setting this is useful when analyzing logs for applications that
                                 are both multi-threaded and use `MultiprocessingPlugins`, defaults
                                 to `PluginThreadName.DEFAULT`.
        :param daemon: Whether or not the process should be a daemon process
        :param configure_child_process_logger: A callable that will be run on the child process to
                                               confirgure concurrent logging
        """
        super().__init__(plugin_name=plugin.name)
        self._plugin = plugin
        self._main_thread_name = main_thread_name
        self._daemon = daemon
        self._configure_child_process_logger = configure_child_process_logger

        self._multiprocessing_context = multiprocessing.get_context("spawn")
        self._receiver, self._sender = multiprocessing.Pipe(duplex=False)

        self._proc = None
        self._calling_thread_name = None
        self._return_value = None

    def run(self, *, timeout: Optional[float] = None, **kwargs) -> Any:
        """
        Run a plugin with the provided keyword arguments and returns the result

        When the timeout argument is not present or None, the operation will block until the
        plugin stops.

        :param: A floating-point number of seconds to wait for the plugin to run

        :return: The data that the plugin returned
        """
        self.start(**kwargs)
        self.join(timeout)

        return self.return_value

    def start(self, **kwargs):
        """
        Launch a new process that runs this plugin
        """
        self._calling_thread_name = current_thread().name

        self._proc = self._multiprocessing_context.Process(
            name=self.name, daemon=self._daemon, target=self._run, kwargs=kwargs
        )
        self._proc.start()

    def _run(self, **kwargs):
        self._set_main_thread_name()
        self._configure_child_process_logger()

        return_value = self._plugin.run(**kwargs)
        self._sender.send(return_value)

    def _set_main_thread_name(self):
        if self._main_thread_name == PluginThreadName.DEFAULT:
            # Do nothing, since the process's main thread will be the interpreter's default
            pass
        elif self._main_thread_name == PluginThreadName.CALLING_THREAD:
            current_thread().name = self._calling_thread_name
        else:
            current_thread().name = self._main_thread_name

    def join(self, timeout: Optional[float] = None):
        """
        Wait for this plugin and its parent process to exit

        When the timeout argument is not present or None, the operation will block until the
        plugin stops.

        :param: A floating-point number of seconds to wait for the plugin to run
        """
        if self._proc is None:
            raise AssertionError("can only join a started plugin")

        self._proc.join(timeout)
        if self.is_alive():
            return

        logger.debug(f"{self.name} exited with code {self._proc.exitcode}")

        self._retrieve_return_value()

    def is_alive(self) -> bool:
        """
        Return whether the plugin is alive (process is still running)

        :return: True if the process/plugin is running. False otherwise.
        """
        if self._proc is None:
            return False

        return self._proc.is_alive()

    def _retrieve_return_value(self):
        try:
            if self._receiver.poll():
                self._return_value = self._read_return_value()
            else:
                logger.error(f"{self.name} did not return a value")
        finally:
            logger.debug(f"Closing Pipe to {self.name}")
            self._receiver.close()
            logger.debug(f"Pipe to {self.name} closed")

    def _read_return_value(self) -> Any:
        try:
            return self._receiver.recv()
            logger.debug(f"{self.name} returned: {self.return_value}")
        except EOFError as err:
            logger.error(f"Error retrieving the return value for {self.name}: {err}")

    @property
    def return_value(self) -> Any:
        """
        The return value of the plugin

        This property will be `None` until the plugin finishes running.
        """
        return self._return_value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['serpentarium']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.4.0,<5.0.0']

setup_kwargs = {
    'name': 'serpentarium',
    'version': '0.6.1',
    'description': 'A Python framework for running plugins with conflicting dependencies',
    'long_description': '# serpentarium\nA Python framework for running plugins with conflicting dependencies\n\n## Description\n\nComing soon!\n\n## Installation\n\n`pip install serpentarium`\n\n## Caveats\n\n- This package is highly experimental\n- `import serpentarium` must be the first thing that your code imports so that\n  it can save the state of the interpreter\'s import system before any other\n  imports modify it.\n- MultiprocessingPlugin only works with the "spawn" method (for now). On Linux,\n  you\'ll need to use a multiprocessing Context object with the "spawn" method\n  to generate any Locks, Events, or other synchronization primitives that will\n  be passed to a plugin.\n- SECURITY: This project loads and executes code from files. Do not load or run\n  plugins from untrusted sources.\n\n## Development\n### Pre-commit hooks\n`pre-commit install`\n',
    'author': 'Mike Salvatore',
    'author_email': 'mike.s.salvatore@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/guardicode/serpentarium',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
